/* smallball - compute and print ballistic tables for small arms */

/*
    Copyright (C) 1994, 1996 Ole-Hj. Kristensen ( ohk@edeber.nta.no )

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.'

    (see end of program)

*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <assert.h>


/* General fourth order Runge Kutta */
void 
rungekutta(float *y, float *dydx, int n, float x, float h, float *yout, 
           void (*d) (float x, float *y, float* dxdy))
                        	/* Initial values of y's */
                           	/* Initial values of derivatives */
                      		/* Number of variables to integrate */
                      		/* Current x value */
                      		/* Step size */
                           	/* New values of y's */
                            	/* Pointer to function which computes
				 * derivatives */
{
    static int      oldn = 0;
    int             i;
    static float    xmid, halfstep, h6, *dy_mid, *dy_trial, *y_trial;

/* Allocate new working area only if number of variables change */
    if (n != oldn) {
	if (n != 0) {
	    if (y_trial) free(y_trial);
	    if (dy_trial) free(dy_trial);
	    if (dy_mid) free(dy_mid);
	    oldn = n;
	}
	dy_mid = (float *) calloc(n, sizeof(float));
	dy_trial = (float *) calloc(n, sizeof(float));
	y_trial = (float *) calloc(n, sizeof(float));
    }
/* Start crunching */

    halfstep = h * 0.5;
    h6 = h / 6.0;
    xmid = x + halfstep;

    for (i = 0; i < n; i++)
	y_trial[i] = y[i] + halfstep * dydx[i];

    (*d) (xmid, y_trial, dy_trial);
    for (i = 0; i < n; i++)
	y_trial[i] = y[i] + halfstep * dy_trial[i];

    (*d) (xmid, y_trial, dy_mid);
    for (i = 0; i < n; i++) {
	y_trial[i] = y[i] + h * dy_mid[i];
	dy_mid[i] += dy_trial[i];
    }

    (*d) (x + h, y_trial, dy_trial);
    for (i = 0; i < n; i++)
	yout[i] = y[i] + h6 * (dydx[i] + dy_trial[i] + 2.0 * dy_mid[i]);
}


float         **yval = 0, *xval = 0;

/* Simple-minded driver for the rungekutta integration routine */
void 
rkdumb(float *vstart, int nvar, float x1, float x2, int nstep, 
       void (*d) (float x, float *y, float* dxdy))
                         	/* Number of variables */
                          	/* Number of steps */
                             	/* Initial values */
                       		/* Integrate from x1 */
                       		/* to x2 */
                            	/* User-supplied function to compute
				 * derivatives */
{
    int             i, k;
    float           x, h;
    float          *vcopy, *vout, *dvcopy;

/* Allocate work space */
    vcopy = (float *) calloc(nvar, sizeof(float));
    vout = (float *) calloc(nvar, sizeof(float));
    dvcopy = (float *) calloc(nvar, sizeof(float));

/* Copy initial values */
    for (i = 0; i < nvar; i++) {
	yval[i][0] = vcopy[i] = vstart[i];
    }

    xval[0] = x1;
    x = x1;
    h = (x2 - x1) / nstep;	/* Step size */

    for (k = 0; k < nstep; k++) {

/* Update y's */
	(*d) (x, vcopy, dvcopy);
	rungekutta(vcopy, dvcopy, nvar, x, h, vout, d);

	assert((x + h) != x);	/* Abort if we don't move */

/* Update x */
	x += h;
	xval[k + 1] = x;

/* Copy result */
	for (i = 0; i < nvar; i++) {
	    yval[i][k + 1] = vcopy[i] = vout[i];
	}
    }

/* Free work space */
    free(dvcopy);
    free(vout);
    free(vcopy);
}

/*
  All units are metric.
  The coordinate system is:
  X is along the line of sight.
  Y is up.
  Z is crosswise.
  */

/*  General form of drag function*/
#define DRAGFUN(vel,a,m) (a * pow(vel,m))

/*
  Drag is dependent on speed in a non-obvious way.
  This is the result of the British 1909 firings.
  Converted to metric units by OHK.
  Other drag functions may be substituted at will

*/
double 
resist_b(double v)
{
    if (v > 7.9248E+02)
	return DRAGFUN(v, 3.406205170590E-03, 1.67);
    if (v > 6.096E+02)
	return DRAGFUN(v, 1.059525135209E-02, 1.5);
    if (v > 4.45008E+02)
	return DRAGFUN(v, 1.547358255244E-03, 1.8);
    if (v > 3.62712E+02)
	return DRAGFUN(v, 1.026963165037E-06, 3.0);
    if (v > 3.16992E+02)
	return DRAGFUN(v, 1.517245393852E-15, 6.45);
    if (v > 2.56032E+02)
	return DRAGFUN(v, 6.451780264672E-07, 3.0);
    return DRAGFUN(v, 1.518071962790E-03, 1.6);
}

/*
   These are the Mayevski functions, which is the basis for Ingalls' tables.
   Converted to US units by Ingalls, and back to metric by OHK.
*/

double 
resist_m(double v)
{
    if (v > 7.924800000000E+02)
	return DRAGFUN(v, 7.813394824668E-03, 1.55);
    if (v > 5.486400000000E+02)
	return DRAGFUN(v, 2.866733530400E-03, 1.7);
    if (v > 4.175760000000E+02)
	return DRAGFUN(v, 4.317622192537E-04, 2.0);
    if (v > 3.749040000000E+02)
	return DRAGFUN(v, 1.030083367220E-06, 3.0);
    if (v > 2.956560000000E+02)
	return DRAGFUN(v, 7.341948762102E-12, 5.0);
    if (v > 2.407920000000E+02)
	return DRAGFUN(v, 6.388708526449E-07, 3.0);
    return DRAGFUN(v, 1.534180425570E-04, 2.0);
}


static int      NSTEP = 1000;	/* Number of steps for integration per second */

typedef struct {
    float           Wx;		/* Wind speed in X, Y and Z */
    float           Wy;
    float           Wz;
    float           bc;		/* Ballistic coefficient */
    double          (*drag) (double);	/* Pointer to drag function */
}               flightparam;

static flightparam fp = {0.0, 0.0, 0.0, 1.0, 0};

/*
  This is our flight model, expressed as a system of differential
  equations. It is rather simple, but should be sufficient for small
  arms fire. It is assumed that the projectile instantaneously adjusts
  itself to the wind, so that it is always pointing directly into the
  relative wind. This makes it simple to compute the effect of the
  crosswind, as the acceleration caused by the wind will just be
  a component of the drag. The drag is computed from the speed of
  the projectile relative to the wind, of course.

  g is assumed to be constant.
  Coriolis forces are not taken into account.

  If the last two assumptions do not hold for you,
  you are into artillery, not small arms fire.
  */

void 
flightmodel(float x, float *y, float *dydx)
{
/* Compute speed relative to wind */
    float           u = sqrt((y[0] + fp.Wx) * (y[0] + fp.Wx) +
			     (y[2] + fp.Wy) * (y[2] + fp.Wy) +
			     (y[4] + fp.Wz) * (y[4] + fp.Wz));

/* Drag components (yes, it really is this simple) */
    float           alfa = (y[0] - fp.Wx) / u;
    float           beta = (y[2] - fp.Wy) / u;
    float           gamma = (y[4] - fp.Wz) / u;

/* Compute drag */
    float           r = fp.drag(u) / fp.bc;

/* Compute acceleration and velocity */
    dydx[0] = -r * alfa;	/* Vx */
    dydx[1] = y[0];		/* Sx */
    dydx[2] = -r * beta - 9.818;/* Vy */
    dydx[3] = y[2];		/* Sy */
    dydx[4] = -r * gamma;	/* Vz */
    dydx[5] = y[4];		/* Sz */
}


/* Linear interpolation is close enough */
float 
linint(float x0, float y0, float x1, float y1, float x)
{
    return (x - x0) * (y1 - y0) / (x1 - x0) + y0;
}

/* Params for the simulation */
typedef struct {
    float           v0;		/* Muzzle velocity */
    float           bc;		/* Ballistic coefficient */
    float           maxtime;	/* Run simulation for maxtime seconds */
    float           maxrange;	/* Display results up to maxrange meters */
    float           sheight;	/* Height of sight above the bore */
    float           crosswind;	/* Speed of wind at right angles to flight */
    float           angle;	/* Elevation */
    float           zero;	/* Where to zero the sights */
    double          (*drag) (double);	/* Drag function */
}               calcparam;


/* Display parameters */
void 
printpar(calcparam p)
{
    printf("# Maximum range is %f m\n", p.maxrange);
    printf("# Zero range is %f m\n", p.zero);
    printf("# Max time of flight is %f s\n", p.maxtime);
    printf("# Ballistic coefficient is %f\n", p.bc);
    printf("# V0 is %f m/s\n", p.v0);
    printf("# Elevation is %f radians (%f degrees) \n", p.angle, p.angle / (2 * M_PI) * 360.0);
    printf("# Height of sight is %f m\n", p.sheight);
    printf("# Crosswind is %f m/s\n", p.crosswind);
}

/* Display result */
void 
printtable(calcparam p)
{
    int             i, j;
    float           t,V,Vx,Vy,Vz,x,y,z;

    printf("#\n# %10.10s %10.10s %10.10s %10.10s %10.10s %10.10s %10.10s %10.10s\n",
	   "t    ", "V    ", "Vx    ", "X    ", "Vy    ", "Y    ", "Vz    ", "Z    ");
    for (i = 1, j = 0; i <= p.maxrange; i += 1) {
	while (j < NSTEP && yval[1][j] < i)
	    j++;
	t = linint(yval[1][j - 1], xval[j - 1], yval[1][j], xval[j], (float) i);

       Vx = linint(xval[j - 1], yval[0][j - 1], xval[j], yval[0][j], t);
       x = linint(xval[j - 1], yval[1][j - 1], xval[j], yval[1][j], t);
       Vy = linint(xval[j - 1], yval[2][j - 1], xval[j], yval[2][j], t);
       y = linint(xval[j - 1], yval[3][j - 1], xval[j], yval[3][j], t);
       Vz = linint(xval[j - 1], yval[4][j - 1], xval[j], yval[4][j], t);
       z = linint(xval[j - 1], yval[5][j - 1], xval[j], yval[5][j], t);
       V = sqrt(Vx*Vx + Vy*Vy + Vz*Vz);

	printf("%10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
	       t,V,Vx,x,Vy,y,Vz,z);

	if (j >= NSTEP)
	    break;
    }
    printf("#%c", 'L' - '@');
}

/* Crude calculation of elevation, given a desired zero range */
void 
zeroin(calcparam *p, float *tval, float *vx, float *vy)
{
    int             i;
    float           t;
    float           dx, dy;

    for (i = 1; i < NSTEP && yval[1][i] < p->zero; i++);
    t = linint(yval[1][i - 1], tval[i - 1], yval[1][i], tval[i], p->zero);
    dx = linint(tval[i - 1], yval[1][i - 1], tval[i], yval[1][i], t);
    dy = linint(tval[i - 1], yval[3][i - 1], tval[i], yval[3][i], t);
    p->angle += atan(-dy / dx);
    *vy = sin(p->angle) * p->v0;
    *vx = cos(p->angle) * p->v0;
}

/*
  Initialize the system and compute tables. The integration of the
  differential equations is done with a 4. order Runge-Kutta method,
  because it is robust and fairly accurate.
  The adventurous may use a Burlisch-Stoer or predictor-corrector
  method if they wish, it should be able to give the same accuracy
  with less computer time.
  */


#define NVAR 6

void 
compute(calcparam p)
{
    float           x1, *vstart;
    int             i;

    x1 = 0.0;			/* Start time at 0 */

    NSTEP = (p.maxtime - x1) * 1000;	/* Step once per millisecond */

    vstart = (float *) calloc(NVAR, sizeof(float));
    /* Note: The arrays xval and yval must have indices up to NSTEP */
    xval = (float *) calloc(NSTEP + 1, sizeof(float));	/* Vector */

    yval = (float **) calloc(NVAR, sizeof(float *));	/* Matrix */
    for (i = 0; i < NVAR; i++)
	yval[i] = (float *) calloc(NSTEP + 1, sizeof(float));

    vstart[0] = p.v0 * cos(p.angle);
    vstart[1] = 0.0;
    vstart[2] = p.v0 * sin(p.angle);
    vstart[3] = -p.sheight;
    vstart[4] = 0.0;
    vstart[5] = 0.0;
    fp.Wz = p.crosswind;
    fp.drag = p.drag;
    fp.bc = p.bc;

    rkdumb(vstart, NVAR, x1, p.maxtime, NSTEP, flightmodel);

    printpar(p);
    printtable(p);

    /* Zero it twice, as the approximation is crude, but converges */
    zeroin(&p, xval, &vstart[0], &vstart[2]);
    rkdumb(vstart, NVAR, x1, p.maxtime, NSTEP, flightmodel);

    zeroin(&p, xval, &vstart[0], &vstart[2]);
    rkdumb(vstart, NVAR, x1, p.maxtime, NSTEP, flightmodel);

    printpar(p);
    printtable(p);

    for (i = 0; i < NVAR; i++)
	free(yval[i]);
    free(yval);
    free(xval);
    free(vstart);
    exit(0);
}

/* Default parameters for simulation */
static calcparam calc = {1000.0, 0.4, 1.0, 1000.0, 0.0, 4.5, 0.0, 100.0, resist_b};

void 
usage(char *name)
{
    printf("Usage: %s\t[-a elevation (radians)] [-c ballistic coefficient]\n", name);
    printf("         \t[-h height of sight] [-r max range] [-t max time of flight]\n");
    printf("         \t[-v velocity] [-w crosswind] [-z zero range]\n");
    printf("         \t[-i] Ingalls tables, British are default\n");
    printf("Please observe that all units are metric\n");
}

main(int argc, char **argv)
{
    extern char    *optarg;
    extern int      optind;
    extern int      getopt(int, char *const *, const char *);
    int             c;
    float           farg;

    while ((c = getopt(argc, argv, "a:c:h:ir:t:v:w:z:")) != -1)
	switch (c) {
	case 'i':
	    calc.drag = resist_m;
	    break;
	case 'z':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.zero = farg;
	    break;
	case 'h':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.sheight = farg;
	    break;
	case 'a':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.angle = farg;
	    break;
	case 'v':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.v0 = farg;
	    break;
	case 'c':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.bc = farg;
	    break;
	case 't':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.maxtime = farg;
	    break;
	case 'r':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.maxrange = farg;
	    break;
	case 'w':
	    if (sscanf(optarg, "%f", &farg) == 1)
		calc.crosswind = farg;
	    break;
	default:
	    usage(argv[0]);
	    exit(1);
	    break;
	}
    compute(calc);
}

/*
		    GNU GENERAL PUBLIC LICENSE
		       Version 2, June 1991

 Copyright (C) 1989, 1991 Free Software Foundation, Inc.
                          675 Mass Ave, Cambridge, MA 02139, USA
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

			    Preamble

  The licenses for most software are designed to take away your
freedom to share and change it.  By contrast, the GNU General Public
License is intended to guarantee your freedom to share and change free
software--to make sure the software is free for all its users.  This
General Public License applies to most of the Free Software
Foundation's software and to any other program whose authors commit to
using it.  (Some other Free Software Foundation software is covered by
the GNU Library General Public License instead.)  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
this service if you wish), that you receive source code or can get it
if you want it, that you can change the software or use pieces of it
in new free programs; and that you know you can do these things.

  To protect your rights, we need to make restrictions that forbid
anyone to deny you these rights or to ask you to surrender the rights.
These restrictions translate to certain responsibilities for you if you
distribute copies of the software, or if you modify it.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must give the recipients all the rights that
you have.  You must make sure that they, too, receive or can get the
source code.  And you must show them these terms so they know their
rights.

  We protect your rights with two steps: (1) copyright the software, and
(2) offer you this license which gives you legal permission to copy,
distribute and/or modify the software.

  Also, for each author's protection and ours, we want to make certain
that everyone understands that there is no warranty for this free
software.  If the software is modified by someone else and passed on, we
want its recipients to know that what they have is not the original, so
that any problems introduced by others will not reflect on the original
authors' reputations.

  Finally, any free program is threatened constantly by software
patents.  We wish to avoid the danger that redistributors of a free
program will individually obtain patent licenses, in effect making the
program proprietary.  To prevent this, we have made it clear that any
patent must be licensed for everyone's free use or not licensed at all.

  The precise terms and conditions for copying, distribution and
modification follow.

		    GNU GENERAL PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
a notice placed by the copyright holder saying it may be distributed
under the terms of this General Public License.  The "Program", below,
refers to any such program or work, and a "work based on the Program"
means either the Program or any derivative work under copyright law:
that is to say, a work containing the Program or a portion of it,
either verbatim or with modifications and/or translated into another
language.  (Hereinafter, translation is included without limitation in
the term "modification".)  Each licensee is addressed as "you".

Activities other than copying, distribution and modification are not
covered by this License; they are outside its scope.  The act of
running the Program is not restricted, and the output from the Program
is covered only if its contents constitute a work based on the
Program (independent of having been made by running the Program).
Whether that is true depends on what the Program does.

  1. You may copy and distribute verbatim copies of the Program's
source code as you receive it, in any medium, provided that you
conspicuously and appropriately publish on each copy an appropriate
copyright notice and disclaimer of warranty; keep intact all the
notices that refer to this License and to the absence of any warranty;
and give any other recipients of the Program a copy of this License
along with the Program.

You may charge a fee for the physical act of transferring a copy, and
you may at your option offer warranty protection in exchange for a fee.

  2. You may modify your copy or copies of the Program or any portion
of it, thus forming a work based on the Program, and copy and
distribute such modifications or work under the terms of Section 1
above, provided that you also meet all of these conditions:

    a) You must cause the modified files to carry prominent notices
    stating that you changed the files and the date of any change.

    b) You must cause any work that you distribute or publish, that in
    whole or in part contains or is derived from the Program or any
    part thereof, to be licensed as a whole at no charge to all third
    parties under the terms of this License.

    c) If the modified program normally reads commands interactively
    when run, you must cause it, when started running for such
    interactive use in the most ordinary way, to print or display an
    announcement including an appropriate copyright notice and a
    notice that there is no warranty (or else, saying that you provide
    a warranty) and that users may redistribute the program under
    these conditions, and telling the user how to view a copy of this
    License.  (Exception: if the Program itself is interactive but
    does not normally print such an announcement, your work based on
    the Program is not required to print an announcement.)

These requirements apply to the modified work as a whole.  If
identifiable sections of that work are not derived from the Program,
and can be reasonably considered independent and separate works in
themselves, then this License, and its terms, do not apply to those
sections when you distribute them as separate works.  But when you
distribute the same sections as part of a whole which is a work based
on the Program, the distribution of the whole must be on the terms of
this License, whose permissions for other licensees extend to the
entire whole, and thus to each and every part regardless of who wrote it.

Thus, it is not the intent of this section to claim rights or contest
your rights to work written entirely by you; rather, the intent is to
exercise the right to control the distribution of derivative or
collective works based on the Program.

In addition, mere aggregation of another work not based on the Program
with the Program (or with a work based on the Program) on a volume of
a storage or distribution medium does not bring the other work under
the scope of this License.

  3. You may copy and distribute the Program (or a work based on it,
under Section 2) in object code or executable form under the terms of
Sections 1 and 2 above provided that you also do one of the following:

    a) Accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of Sections
    1 and 2 above on a medium customarily used for software interchange; or,

    b) Accompany it with a written offer, valid for at least three
    years, to give any third party, for a charge no more than your
    cost of physically performing source distribution, a complete
    machine-readable copy of the corresponding source code, to be
    distributed under the terms of Sections 1 and 2 above on a medium
    customarily used for software interchange; or,

    c) Accompany it with the information you received as to the offer
    to distribute corresponding source code.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form with such
    an offer, in accord with Subsection b above.)

The source code for a work means the preferred form of the work for
making modifications to it.  For an executable work, complete source
code means all the source code for all modules it contains, plus any
associated interface definition files, plus the scripts used to
control compilation and installation of the executable.  However, as a
special exception, the source code distributed need not include
anything that is normally distributed (in either source or binary
form) with the major components (compiler, kernel, and so on) of the
operating system on which the executable runs, unless that component
itself accompanies the executable.

If distribution of executable or object code is made by offering
access to copy from a designated place, then offering equivalent
access to copy the source code from the same place counts as
distribution of the source code, even though third parties are not
compelled to copy the source along with the object code.

  4. You may not copy, modify, sublicense, or distribute the Program
except as expressly provided under this License.  Any attempt
otherwise to copy, modify, sublicense or distribute the Program is
void, and will automatically terminate your rights under this License.
However, parties who have received copies, or rights, from you under
this License will not have their licenses terminated so long as such
parties remain in full compliance.

  5. You are not required to accept this License, since you have not
signed it.  However, nothing else grants you permission to modify or
distribute the Program or its derivative works.  These actions are
prohibited by law if you do not accept this License.  Therefore, by
modifying or distributing the Program (or any work based on the
Program), you indicate your acceptance of this License to do so, and
all its terms and conditions for copying, distributing or modifying
the Program or works based on it.

  6. Each time you redistribute the Program (or any work based on the
Program), the recipient automatically receives a license from the
original licensor to copy, distribute or modify the Program subject to
these terms and conditions.  You may not impose any further
restrictions on the recipients' exercise of the rights granted herein.
You are not responsible for enforcing compliance by third parties to
this License.

  7. If, as a consequence of a court judgment or allegation of patent
infringement or for any other reason (not limited to patent issues),
conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot
distribute so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you
may not distribute the Program at all.  For example, if a patent
license would not permit royalty-free redistribution of the Program by
all those who receive copies directly or indirectly through you, then
the only way you could satisfy both it and this License would be to
refrain entirely from distribution of the Program.

If any portion of this section is held invalid or unenforceable under
any particular circumstance, the balance of the section is intended to
apply and the section as a whole is intended to apply in other
circumstances.

It is not the purpose of this section to induce you to infringe any
patents or other property right claims or to contest validity of any
such claims; this section has the sole purpose of protecting the
integrity of the free software distribution system, which is
implemented by public license practices.  Many people have made
generous contributions to the wide range of software distributed
through that system in reliance on consistent application of that
system; it is up to the author/donor to decide if he or she is willing
to distribute software through any other system and a licensee cannot
impose that choice.

This section is intended to make thoroughly clear what is believed to
be a consequence of the rest of this License.

  8. If the distribution and/or use of the Program is restricted in
certain countries either by patents or by copyrighted interfaces, the
original copyright holder who places the Program under this License
may add an explicit geographical distribution limitation excluding
those countries, so that distribution is permitted only in or among
countries not thus excluded.  In such case, this License incorporates
the limitation as if written in the body of this License.

  9. The Free Software Foundation may publish revised and/or new versions
of the General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

Each version is given a distinguishing version number.  If the Program
specifies a version number of this License which applies to it and "any
later version", you have the option of following the terms and conditions
either of that version or of any later version published by the Free
Software Foundation.  If the Program does not specify a version number of
this License, you may choose any version ever published by the Free Software
Foundation.

  10. If you wish to incorporate parts of the Program into other free
programs whose distribution conditions are different, write to the author
to ask for permission.  For software which is copyrighted by the Free
Software Foundation, write to the Free Software Foundation; we sometimes
make exceptions for this.  Our decision will be guided by the two goals
of preserving the free status of all derivatives of our free software and
of promoting the sharing and reuse of software generally.

			    NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.

		     END OF TERMS AND CONDITIONS

*/
